from __future__ import absolute_import

from datetime import datetime, timedelta

import requests
from six import string_types

from ...utilities import timezone
from ..exceptions import InvalidArgument

try:
    # noinspection PyUnresolvedReferences
    from typing import Union, Any
except ImportError:
    pass


_valid_identifier_characters = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._')
_min_identifier_length = 1
_max_identifier_length = 1024

_allowed_file_types = {
    # PDFs
    'pdf',
    # Word documents
    'docx', 'docm', 'doc', 'rtf',
    # PowerPoint presentations
    'pptx', 'pptm', 'ppt',
}


def validate_identifier(identifier):
    # type: (str) -> str
    if not identifier:
        raise InvalidArgument('identifier', '`identifier` cannot be empty.')
    if not isinstance(identifier, string_types):
        raise InvalidArgument('identifier', '`identifier` must be a string (of type `str`).')
    if not _min_identifier_length <= len(identifier) <= _max_identifier_length:
        raise InvalidArgument('identifier', '`identifier` must be between {min} and {max} characters long.'.format(min=_min_identifier_length, max=_max_identifier_length))
    if not all(c in _valid_identifier_characters for c in identifier):
        raise InvalidArgument('identifier', '`identifier` must only contain ASCII letters, numbers, and the characters "-._".')
    return identifier


def validate_file_type(file_type):
    # type: (str) -> str
    if not file_type:
        raise InvalidArgument('file_type', '`file_type` cannot be empty.')
    munged_file_type = str(file_type).lower()
    if munged_file_type not in _allowed_file_types:
        raise InvalidArgument('file_type', '"{}" is not a valid file type'.format(file_type))
    return munged_file_type


def validate_file(file):
    # type: (Any) -> Any
    if not file:
        raise InvalidArgument('file', '`file` must be specified.')
    if any(not hasattr(file, attr) for attr in ('mode', 'read')):
        raise InvalidArgument('file', "the given object doesn't appear to be a file - it's missing properties a file object should have.")
    if hasattr(file, 'readable') and not file.readable():
        raise InvalidArgument('file', '`file.readable()` is False.')
    if 'b' not in file.mode:
        raise InvalidArgument('file', "the given file doesn't appear to be open in binary mode. It's mode is '{}'.".format(file.mode))
    return file


def validate_url(url):
    # type: (str) -> str
    if not url:
        raise InvalidArgument('url', '`url` cannot be empty.')
    if not isinstance(url, string_types):
        raise InvalidArgument('url', '`url` must be a string (of type `str`).')
    if len(url) > 2048:
        raise InvalidArgument('url', "`url` must be no longer than 2048 characters.")
    try:
        scheme, netloc, _, _, _, _ = requests.utils.urlparse(url)
    except Exception as ex:
        wrapper = InvalidArgument('url', 'the value could not be parsed as a URL.')
        wrapper.__cause__ = ex
        raise wrapper
    if scheme not in ('http', 'https'):
        raise InvalidArgument('url', '`url` must have "http" or "https" as its scheme.')
    if not netloc:
        raise InvalidArgument('url', 'the given URL is malformed.')
    return url


def _validate_datetime_or_timedelta(parameter_name, value):
    # type: (Union[datetime, timedelta]) -> datetime
    if isinstance(value, timedelta):
        if value.total_seconds() <= 0:
            raise InvalidArgument(parameter_name, 'if a timedelta, `{parameter}` must be positive, but it was zero or negative.'.format(parameter=parameter_name))
        return datetime.now(tz=timezone.utc) + value

    if not isinstance(value, datetime):
        raise InvalidArgument(parameter_name, '`{parameter}` must be a datetime or a timedelta.'.format(parameter=parameter_name))
    # Make the datetime aware, with UTC as its timezone. If it was naive, we assume it was in UTC time.
    if value.utcoffset() is None:
        value = value.replace(tzinfo=timezone.utc)
    else:
        value = value.astimezone(timezone.utc)
    # Now we can safely check if it's in the future.
    if value < datetime.now(tz=timezone.utc):
        raise InvalidArgument(parameter_name, '`{parameter}` must be in the future.'.format(parameter=parameter_name))

    return value


def validate_expires(expires):
    # type: (Union[datetime, timedelta]) -> datetime
    return _validate_datetime_or_timedelta('expires', expires)


def validate_valid_until(valid_until):
    # type: (Union[datetime, timedelta]) -> datetime
    return _validate_datetime_or_timedelta('valid_until', valid_until)



from __future__ import absolute_import

import functools

import requests

try:
    # noinspection PyUnresolvedReferences
    from typing import Union
except ImportError:
    pass


class EndpointException(Exception):
    pass


class InvalidPath(Exception):
    pass


class InvalidArgument(EndpointException):
    def __init__(self, argument_name, message):
        # type: (str, str) -> None
        super(InvalidArgument, self).__init__("An invalid value was received for `{argument}`: {message}".format(argument=argument_name, message=message))


class BadRequest(EndpointException):
    def __init__(self, status_code, response):
        # type: (int, Union[dict, list]) -> None
        self.status_code = status_code
        self.response = response
        super(BadRequest, self).__init__('Bad request: status={status}, response={response}'.format(status=status_code, response=response))


class NotFound(BadRequest):
    pass


def raise_for(ex):
    # type: (requests.exceptions.RequestException) -> None
    if isinstance(ex, requests.exceptions.HTTPError):
        try:
            if ex.response.status_code == 404:
                wrapper = NotFound(404, ex.response.json())
            else:
                wrapper = BadRequest(ex.response.status_code, ex.response.json())
        except:
            # No JSON body? This shouldn't happen, but we'll rethrow anyway.
            wrapper = BadRequest(ex.response.status_code, ex.response)
    else:
        # An error in communication has occurred.
        wrapper = EndpointException("Unable to connect to the API. This can occur when you attempt to upload files, "
                                    "but have invalid credentials - please check your credentials.")

    wrapper.__cause__ = ex
    raise wrapper


def handle_request_exception(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except requests.exceptions.RequestException as ex:
            raise_for(ex)

    return wrapper



from __future__ import absolute_import

from .endpoints import ComparisonsEndpoint
from .transport import RESTClient
from .utilities.urls import Url

try:
    from typing import Optional
except ImportError:
    pass


PRODUCTION_CLOUD_BASE_URL = 'https://api.draftable.com/v1'


class Client(object):
    def __init__(self, account_id, auth_token, base_url=None):
        # type: (str, str, Optional[str]) -> None
        self.__client = RESTClient(account_id, auth_token)
        self.__base_url = Url(base_url or PRODUCTION_CLOUD_BASE_URL)
        self.comparisons = ComparisonsEndpoint(self.__client, self.__base_url)

    @property
    def account_id(self):
        # type: () -> str
        return self.__client.account_id

    @property
    def auth_token(self):
        # type: () -> str
        return self.__client.auth_token

    @property
    def base_url(self):
        # type: () -> str
        return str(self.__base_url)

    @property
    def verify_ssl(self):
        # type: () -> bool
        return str(self.__client.verify_ssl)

    @verify_ssl.setter
    def verify_ssl(self, v):
        # type: (v) -> None
        self.__client.verify_ssl = v

    def __repr__(self):
        # type: () -> str
        return 'Client(account_id={}, auth_token={}, base_url={})'.format(repr(self.account_id), repr(self.auth_token), repr(self.base_url))


Client.__str__ = Client.__repr__



from __future__ import absolute_import

from os.path import basename, isfile, join, splitext

from six import string_types
# Note: `urllib3` is a required dependency of `requests`
from urllib3.util import parse_url

from ..exceptions import InvalidPath
from . import validation

try:
    # noinspection PyUnresolvedReferences
    from typing import List, Union, Optional, Any
except ImportError:
    pass


class Side(object):
    def __init__(self, file_type, display_name=None):
        # type: (str, Optional[str]) -> None
        self.__file_type = validation.validate_file_type(file_type)
        self.__display_name = None if display_name is None else str(display_name)

    @property
    def file_type(self):
        # type: () -> str
        return self.__file_type

    @property
    def display_name(self):
        # type: () -> Optional[str]
        return self.__display_name


class FileSide(Side):
    # noinspection PyShadowingBuiltins
    def __init__(self, file, file_type, display_name=None):
        # type: (Any, str, Optional[str]) -> None
        self.__file = validation.validate_file(file)
        super(FileSide, self).__init__(file_type=file_type, display_name=display_name)

    @property
    def file(self):
        # type: () -> Any
        return self.__file

    def __str__(self):
        return "File side: {} ({}, '{}')".format(self.__file, self.file_type, self.display_name)


class URLSide(Side):
    def __init__(self, url, file_type, display_name=None):
        # type: (str, str, Optional[str]) -> None
        self.__url = validation.validate_url(url)
        super(URLSide, self).__init__(file_type=file_type, display_name=display_name)

    @property
    def url(self):
        # type: () -> str
        return self.__url

    def __str__(self):
        return "URL side: {} ({}, '{}')".format(self.__url, self.file_type, self.display_name)


def data_from_side(side_name, side):
    """Given a side (which may be a string or Side object, return the data needed
    for a comparison POST request as a dictionary.
    """
    # type: (str, Union[str, FileSide, URLSide]) -> dict

    if isinstance(side, string_types):
        # User has provided a file path or URL.
        side = make_side(side)

    data = {
        'file_type': side.file_type,
    }

    if side.display_name:
        data['display_name'] = side.display_name
    if isinstance(side, URLSide):
        data['source_url'] = side.url
    else:
        assert isinstance(side, FileSide)
        data['file'] = ("{}.{}".format(side_name, side.file_type), side.file, join('application', 'octet-stream'))
    return data


def side_from_url(url, file_type, display_name=None):
    # type: (str, str, Optional[str]) -> URLSide
    return URLSide(url, file_type, display_name)


def side_from_file(file_obj, file_type, display_name=None):
    # type: (Any, str, Optional[str]) -> FileSide
    return FileSide(file_obj, file_type, display_name)


def side_from_file_path(file_path, file_type=None, display_name=None):
    should_guess = file_type is None or file_type == 'guess'
    if should_guess:
        file_type = guess_file_type_from_path(file_path)
    display_name = display_name or basename(file_path)

    # TODO: make sure this closes the file handle
    return FileSide(open(file_path, 'rb'), file_type, display_name)


def make_side(url_or_file_path, file_type=None, display_name=None):
    """
    Parsing the URL or file path and looking for the file extension is "ok"
    but works only based on human conventions. The more industrial strength
    approach is to read the beginning of the file or URL and determine the
    type unambiguously from the content.

    :param url_or_file_path: a URL or file path to compare.
    :param file_type: a string like "pdf", "docx", etc, or "guess" (or None) to detect.
    :return: a Side object
    """
    # type: (str, Optional[str])
    guess_type = file_type is None or file_type == 'guess'

    if url_or_file_path.startswith("http"):
        url = parse_url(url_or_file_path)  # parse to get the path component on its own, without query string
        if guess_type:
            file_type = guess_file_type_from_path(url.path)
        display_name = display_name or basename(url.path)
        return URLSide(url.url, file_type, display_name)

    elif url_or_file_path.startswith("file:"):
        url = parse_url(url_or_file_path)
        if url.host:
            raise InvalidPath("File url '{}' must be local only, and not contain host name ('{}')".format(
                url_or_file_path, url.host))

        if not isfile(url.path):
            raise InvalidPath("File url '{}' refers to file '{}' but no such file exists".format(
                url_or_file_path, url.path))

        return side_from_file_path(url.path, file_type, display_name)

    elif isfile(url_or_file_path):
        return side_from_file_path(url_or_file_path, file_type, display_name)

    else:
        raise ValueError("Path is not a URL and not a file that exists: {}".format(url_or_file_path))


def guess_file_type_from_path(path):
    _, file_type = splitext(path)
    return file_type.strip('.').lower()


def guess_file_type_from_url(url):
    url = parse_url(url)  # parse to get the path component on its own, without query string
    file_type = guess_file_type_from_path(url.path)
    return file_type



from __future__ import absolute_import

import requests

from ..utilities import Url

try:
    # noinspection PyUnresolvedReferences
    from typing import Any, Union, Tuple, Optional
except ImportError:
    pass


def _is_file(obj):
    # type: (Any) -> bool
    # This is how the requests library checks.
    return hasattr(obj, '__iter__') and not isinstance(obj, (str, list, tuple, dict))


def _is_file_tuple(obj):
    # type: (Any) -> bool
    return isinstance(obj, tuple) and any(map(_is_file, obj))


def _data_contains_file(data):
    # type: (Any) -> bool
    if isinstance(data, (list, tuple)):
        return any(map(_data_contains_file, data))

    if isinstance(data, dict):
        return any(map(_data_contains_file, data.values()))

    return _is_file(data)


# Flattens our nested form data, e.g. "right: {file_type: ...}" becomes "right.file_type = ...".
# Additionally, separates plain data from file objects.
# This lets us support uploading files "nested" in a subkey, by setting the key to e.g. "left.file".
# At least, Django Rest Framework is happy to receive data in this format.
def _flatten_form_data(initial_data):
    # type: (dict) -> Tuple[dict, dict]
    data = dict()
    files = dict()

    for key, value in initial_data.items():
        if isinstance(value, dict):
            flattened_data, flattened_files = _flatten_form_data(value)
            for sub_key, sub_value in flattened_data.items():
                data['{}.{}'.format(key, sub_key)] = sub_value
            for sub_key, sub_file in flattened_files.items():
                files['{}.{}'.format(key, sub_key)] = sub_file
        else:
            if _is_file(value) or _is_file_tuple(value):
                files[key] = value
            else:
                data[key] = value

    return data, files


class RESTClient(object):
    def __init__(self, account_id, auth_token):
        # type: (str, str) -> None
        self.__account_id = account_id
        self.__auth_token = auth_token
        self.verify_ssl = True

    @property
    def account_id(self):
        # type: () -> str
        return self.__account_id

    @property
    def auth_token(self):
        # type: () -> str
        return self.__auth_token

    def __auth(self, r):
        r.headers['Authorization'] = 'Token {}'.format(self.__auth_token)
        return r

    def get(self, url, parameters=None):
        # type: (Union[str, Url], Optional[dict]) -> Union[dict, list]
        response = requests.get(url, auth=self.__auth, params=parameters, verify=self.verify_ssl)
        response.raise_for_status()
        return response.json()

    def post(self, url, data):
        # type: (str, dict) -> Union[dict, list]
        if not _data_contains_file(data):
            response = requests.post(url, auth=self.__auth, json=data, verify=self.verify_ssl)
        else:
            data, files = _flatten_form_data(data)
            # Obscure issue:
            # When the request is bad (e.g. invalid authentication), requests throws a weird ConnectionError rather than a HTTPError. Only in this multipart case!
            # (It seems that when the request is bad, our API (via Django Rest Framework) may not wait for the full upload?)
            # Asking for JSON seems to help? But it fails with frequency ~30% when you give invalid credentials.
            # I don't have a good fix for this (yet!), so there's a note in the exception thrown in the weird case. ~ James (April 2017)
            response = requests.post(url, auth=self.__auth, data=data, files=files, headers={'Accept': 'application/json'}, verify=self.verify_ssl)

        response.raise_for_status()
        return response.json()

    def delete(self, url):
        # type: (str) -> None
        response = requests.delete(url, auth=self.__auth, verify=self.verify_ssl)
        response.raise_for_status()



from __future__ import absolute_import

# noinspection PyUnresolvedReferences
from .client import PRODUCTION_CLOUD_BASE_URL, Client
from .endpoints.comparisons.identifier import generate_identifier
from .endpoints.comparisons.sides import make_side

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

__version__ = version('draftable-compare-api')



import random
import string

# Constants for generating random unique (with high probability) identifiers:
_randomIdentifierLength = 12
_randomIdentifierCharset = string.ascii_letters


def generate_identifier():
    # type: () -> str
    return ''.join(random.choice(_randomIdentifierCharset) for _ in range(_randomIdentifierLength))



from __future__ import absolute_import

import hashlib
import hmac
import json
from collections import OrderedDict


def get_viewer_url_signature(account_id, auth_token, identifier, valid_until_timestamp):
    # type: (str, str, str, int) -> str

    # First, we generate our policy, which is an object giving the account_id and identifier of the comparison, and the valid_until time as a timestamp.
    policy = OrderedDict((
        ('account_id', str(account_id)),
        ('identifier', str(identifier)),
        ('valid_until', int(valid_until_timestamp)),
    ))

    # Next, we convert the policy to JSON. We've used an OrderedDict so the fields will be serialized in the correct order.
    # Note that we don't have any whitespace in our JSON, achieved by specifying the separators.
    json_policy = json.dumps(policy, separators=(',', ':'))

    # Finally, we compute a SHA-256 HMAC of the JSON policy, using our auth token as the secret key.
    # The API expects it to be hex-encoded.
    return hmac.new(key=auth_token.encode('utf-8'), msg=json_policy.encode('utf-8'), digestmod=hashlib.sha256).hexdigest()



from datetime import datetime

from . import timezone

_unix_epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)


# A Python 2 compatible method for getting timestamps...

def aware_datetime_to_timestamp(dt):
    # type: (datetime) -> int
    return int((dt - _unix_epoch).total_seconds())


def parse_datetime(iso_format_string):
    # type: (str) -> datetime
    try:
        return datetime.strptime(iso_format_string, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
    except ValueError:
        # Sometimes the datetime can be missing the milliseconds, in which case the strptime call fails.
        return datetime.strptime(iso_format_string, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)



#!/usr/bin/env python

from __future__ import print_function

import argparse
import datetime
import os
import sys

from draftable import Client as DraftableClient
from draftable.endpoints.comparisons.sides import make_side
from draftable.endpoints.exceptions import (InvalidArgument, InvalidPath,
                                            NotFound)

try:
    import configparser  # Python 3
except ImportError:
    import ConfigParser as configparser  # Python 2


description = 'Create and manage Draftable.com comparisons on the command line'

usage_template = """%(prog)s <command> [<command-args>]

COMMANDS:

{commands}

CONFIGURATION:

  Both account_id and auth_token must be set. They can be provided several ways:

  1. Environment variables (DR_ACCOUNT, DR_TOKEN)
  2. A named configuration group in file `~/.draftable`, for example:

      [cloud-testing]  # replace the following with your own credentials
      account: F91n2k-test
      token: d9c3b60dcb83a7bb87efc5c8e72a0599

     To use the above, either specify `-e cloud-testing` as flags to this script,
     or `DR_ENV=cloud-testing` as an environment variable.

  3. As command-line flags, for example:

       "-a <account> -t <token>"
     or
       "--account <account> --token <token>".

  The `base_url` is needed only for self-hosted or development purposes. It can be
  set from environment variable (DR_BASE_URL), preset in the configuration file,
  or a command-line flag (`-b <url>`).

  The `base_url` must always end with '/api' with no trailing slash.

EXAMPLES:

  The following assume credentials are in the environment.

  Create a comparison:

    $ dr-compare create ./left.pdf ./right.pdf

  Create a public comparison that expires in 15 minutes:

    $ dr-compare create -p -m 15 ./left.pdf ./right.pdf

  Provide the file types when not clear from the file extension:

    $ dr-compare create --left-type=pdf --right-type=rtf ./left-file ./right-file

  List all comparisons:

    $ dr-compare list
    $ dr-compare all

  Get a specific comparison:

    $ dr-compare get PCiIEXzW

  Get public URL:

    $ dr-compare url PCiIEXzW

  Get signed URL with 30 minute expiry:

    $ dr-compare signed PCiIEXzW

  Get signed URL with 75 minute expiry:

    $ dr-compare signed PCiIEXzW -m 75
    $ dr-compare signed PCiIEXzW --expiry-mins 75
"""


class SetupError(Exception):
    pass


def get_connection_args(name):
    ini_path = os.path.join(os.path.expanduser("~"), ".draftable")
    if not os.path.isfile(ini_path):
        raise SetupError("Requested environment '%s' but missing config file: %s" % (name, ini_path))

    config = configparser.ConfigParser()
    config.read(ini_path)

    # Only Python 3 supports "if name in config"
    if not config.has_section(name):
        raise SetupError("Requested environment '%s' but config file (%s) does not define that name." % (name, ini_path))

    try:
        base_url = config.get(name, 'base_url')  # Python 3 allows `fallback=None`, but not python 2
    except configparser.NoOptionError:
        # no base_url, that's fine
        base_url = None

    return (
        config.get(name, 'account'),
        config.get(name, 'token'),
        base_url,
    )


def create_client(args):
    # First, use any environment variables
    account = os.getenv('DR_ACCOUNT')
    token = os.getenv('DR_TOKEN')
    base_url = os.getenv('DR_BASE_URL')

    # Secondly, if provided, look up this name in the ~/.draftable file (if it exists)
    env_name = getattr(args, 'env_name') or os.getenv('DR_ENV')  # getattr default value doesn't work with args object

    if env_name:
        account, token, base_url = get_connection_args(env_name)

    # Thirdly, allow the user to override both environment vars and predefined settings with command-line flags.
    account = args.account if args.account else account
    token = args.token if args.token else token
    base_url = args.base_url if args.base_url else base_url

    # Some sanity checks
    if not account or not token:
        raise SetupError("Both account and token must be set.")

    client = DraftableClient(account, token, base_url)
    if args.unverified_ssl:
        client.verify_ssl = False
    return client


def with_std_options(arg_parser):
    """Embellish the provided `arg_parser` with standard options.
    """
    arg_parser.add_argument('-a', '--account', metavar='<ACCOUNT-ID>', action='store', help="For cloud, see https://api.draftable.com/account/credentials")
    arg_parser.add_argument('-b', '--base-url', metavar='<BASE-URL>', action='store', help="Only for Enterprise self-hosted")
    arg_parser.add_argument('-e', '--env-name', metavar='<ENV-NAME>', action='store', help="Name from file $HOME/.draftable")
    arg_parser.add_argument('-t', '--token', metavar='<TOKEN>', action='store', help="For cloud, see https://api.draftable.com/account/credentials")
    arg_parser.add_argument('-S', '--unverified-ssl', action='store_true', help="Don't verify SSL validity; useful for self-hosted self-signed SSL")
    return arg_parser


def default_comparison_display(comp, out=sys.stdout, position=None):
    """Print a comparison to stdout
    """
    out.write("Comparison")
    if position is not None:
        out.write(" %s" % position)
    out.write(" identifier: {}\n".format(comp.identifier))
    out.write("  ready:       {}\n".format(comp.ready))
    out.write("  failed:      {}\n".format(comp.failed))
    out.write("  error:       {}\n".format(comp.error_message))
    out.write("  public:      {}\n".format(comp.public))
    out.write("  created:     {}\n".format(comp.creation_time))
    out.write("  expires:     {}\n".format(comp.expiry_time))
    out.write("  left:        {}\n".format(comp.left))
    out.write("  right:       {}\n".format(comp.right))


def create_comparison(system_args, prog, cmd_name):
    """Create a new comparison."""
    arg_parser = with_std_options(argparse.ArgumentParser(description=create_comparison.__doc__))

    arg_parser.add_argument('-p', '--public', action='store_true', default=False, help="Marks this comparison public")
    arg_parser.add_argument('-i', '--identifier', default=None, help="Provide your own comparison ID")
    arg_parser.add_argument('-m', '--expiry-mins', metavar='<MINS>', type=int, default=None, help='number of minutes this URL should be valid')

    arg_parser.add_argument('--left-type', default='guess', metavar='<EXT>', action='store', help="e.g. 'pdf', 'doc', 'docx', 'ppt', 'pptx'.")
    arg_parser.add_argument('--right-type', default='guess', metavar='<EXT>', action='store')
    arg_parser.add_argument('left', metavar='left-file-path-or-url')
    arg_parser.add_argument('right', metavar='right-file-path-or-url')

    # arg_parser.add_argument('--amend', action='store_true')

    args = arg_parser.parse_args(system_args)
    # print('Running create, args:', args)

    client = create_client(args)
    # print("Client:", client)

    try:
        left = make_side(args.left, args.left_type)
        right = make_side(args.right, args.right_type)
    except InvalidArgument as e:
        raise SetupError("{}. You may need to specify file type with --left-type or --right-type".format(e))
    except InvalidPath as e:
        raise SetupError(str(e))

    identifier = args.identifier
    public = args.public
    expires = datetime.timedelta(minutes=args.expiry_mins) if args.expiry_mins else None

    print("Create:")
    print("  identifier: %s" % identifier)
    print("  public: %s" % public)
    print("  expires: %s" % expires)
    print("  left: %s" % left)
    print("  right: %s" % right)
    print("  ...")
    comp = client.comparisons.create(left, right, identifier, public, expires)

    display = default_comparison_display
    display(comp)
    print_basic_urls(client, comp)


def print_basic_urls(client, comp):
    url_expiry = datetime.timedelta(minutes=30)
    print("\nURLs:")
    print("  public URL:  %s" % client.comparisons.public_viewer_url(comp.identifier))
    print("  signed URL:  %s" % client.comparisons.signed_viewer_url(comp.identifier, url_expiry))
    print("     expires:  %s" % url_expiry)


def list_all_comparisons(system_args, prog, cmd_name):
    """Retrieve and display all comparisons."""

    arg_parser = with_std_options(argparse.ArgumentParser(description=list_all_comparisons.__doc__))

    args = arg_parser.parse_args(system_args)
    # print('Running list, args:', args)
    client = create_client(args)
    display = default_comparison_display
    # print("Client:", client)
    comparisons = client.comparisons.all()
    n = len(comparisons)

    print("Account %s has %d comparison(s):" % (client.account_id, n))
    for i, comp in enumerate(comparisons, 1):
        display(comp, position="%d of %d" % (i, n))


def list_one_comparison(system_args, prog, cmd_name):
    """Retrieve and display specific comparison or comparisons."""
    arg_parser = with_std_options(argparse.ArgumentParser(
        prog='%s %s' % (prog, cmd_name),  # so that "-h/--help" shows "dr-compare <cmd>"
        description=list_one_comparison.__doc__))
    arg_parser.add_argument('identifiers', metavar='<ID>', nargs='+', help='a comparison identifier')

    args = arg_parser.parse_args(system_args)
    client = create_client(args)
    display = default_comparison_display
    n = len(args.identifiers)

    for i, identifier in enumerate(args.identifiers, 1):
        try:
            comp = client.comparisons.get(identifier)
            display(comp, position="%d of %d" % (i, n))
            print_basic_urls(client, comp)
        except NotFound as e:
            print("Comparison not found with identifier: %s" % identifier)


def delete_comparison(system_args, prog, cmd_name):
    """Delete a specific comparison."""
    arg_parser = with_std_options(argparse.ArgumentParser(
        prog='%s %s' % (prog, cmd_name),  # so that "-h/--help" shows "dr-compare <cmd>"
        description=list_one_comparison.__doc__))
    arg_parser.add_argument('identifier', metavar='<ID>', help='a comparison identifier')

    args = arg_parser.parse_args(system_args)
    client = create_client(args)
    identifier = args.identifier

    try:
        client.comparisons.delete(identifier)
    except NotFound as e:
        print("Comparison not found with identifier: %s" % identifier)


def show_public_url(system_args, prog, cmd_name):
    """Generate a public (unsigned) URL to view a specific comparison."""
    arg_parser = with_std_options(argparse.ArgumentParser(
        prog='%s %s' % (prog, cmd_name),  # so that "-h/--help" shows "dr-compare <cmd>"
        description=list_one_comparison.__doc__))
    arg_parser.add_argument('identifiers', metavar='<ID>', nargs='+', help='a comparison identifier')

    args = arg_parser.parse_args(system_args)
    client = create_client(args)

    for i, identifier in enumerate(args.identifiers, 1):
        print(client.comparisons.public_viewer_url(identifier))


def show_signed_url(system_args, prog, cmd_name):
    """Generate a signed URL to view a specific comparison."""
    arg_parser = with_std_options(argparse.ArgumentParser(
        prog='%s %s' % (prog, cmd_name),  # so that "-h/--help" shows "dr-compare <cmd>"
        description=list_one_comparison.__doc__))
    arg_parser.add_argument('-m', '--expiry-mins', metavar='<MINS>', type=int, default=30, help='number of minutes this URL should be valid')
    arg_parser.add_argument('identifiers', metavar='<ID>', nargs='+', help='a comparison identifier')

    args = arg_parser.parse_args(system_args)
    client = create_client(args)
    url_expiry = datetime.timedelta(minutes=args.expiry_mins)

    for i, identifier in enumerate(args.identifiers, 1):
        print(client.comparisons.signed_viewer_url(identifier, url_expiry))


commands = dict(
    create=create_comparison,
    all=list_all_comparisons,
    get=list_one_comparison,
    delete=delete_comparison,
    url=show_public_url,
    signed=show_signed_url,
)

aliases = {
    'new': 'create',
    'post': 'create',
    'add': 'create',
    'list': 'all',
    'getall': 'all',
    'del': 'delete',
    'rm': 'delete',
    'public': 'url',
    'public-url': 'url',
    'public-url': 'url',
    'signed-url': 'signed',
    'signed_url': 'signed',
}


def make_usage(template, command_map, alias_map):
    """Generate the usage doc based on configured commands and aliases
    """
    def format_command_info(command_name):
        func = command_map[command_name]

        # Some commands (but not all) have aliases
        a = [k for k in alias_map.keys() if alias_map[k]==command_name]
        a = '\n           Aliases: %s' % ' '.join(sorted(a)) if a else ''
        return "  {:8s} {}{}\n".format(command_name, func.__doc__, a)

    command_info_parts = map(format_command_info, (name for name in sorted(command_map.keys())))
    return template.format(commands='\n'.join(command_info_parts))


def dr_compare_main(system_args=None):
    if system_args is None:
        system_args = sys.argv

    arg_parser = argparse.ArgumentParser(description=description, usage=make_usage(usage_template, commands, aliases))
    arg_parser.add_argument('command', help='Command to run')
    args = arg_parser.parse_args(system_args[1:2])  # just to select the command

    if args.command == 'help':
        arg_parser.print_usage()
        sys.exit(0)

    command_name = aliases.get(args.command, args.command)
    command = commands.get(command_name)

    if not command:
        err = "Invalid command '%s'. For list of commands: %s -h\n" % (command_name, arg_parser.prog)
        arg_parser.error(err)

    command(system_args[2:], arg_parser.prog, command_name)


if __name__ == '__main__':
    try:
        dr_compare_main(sys.argv)
    except SetupError as e:
        sys.stderr.write("Error: %s\n" % e)
        sys.exit(1)



from __future__ import absolute_import

try:
    # noinspection PyUnresolvedReferences
    from datetime import timezone
    utc = timezone.utc

except ImportError:

    # If we're in python 2, we need to implement our own tzinfo class in order to make datetimes be aware...

    from datetime import timedelta, tzinfo

    ZERO = timedelta(0)

    # A UTC class.

    class UTC(tzinfo):
        """UTC"""

        def utcoffset(self, dt):
            return ZERO

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return ZERO

    utc = UTC()



from __future__ import absolute_import, division

from datetime import datetime, timedelta

from ...transport import RESTClient
from ...utilities import Url, aware_datetime_to_timestamp
from ..exceptions import handle_request_exception
from . import signing
from .comparison import Comparison, comparison_from_response
from .sides import FileSide, URLSide, data_from_side
from .validation import (validate_expires, validate_identifier,
                         validate_valid_until)

try:
    # noinspection PyUnresolvedReferences
    from typing import List, Union, Optional, Any
except ImportError:
    pass


class ComparisonsEndpoint(object):

    def __init__(self, client, base_url):
        # type: (RESTClient, Url) -> None
        self.__url = base_url / 'comparisons'
        self.__client = client

    @property
    def account_id(self):
        # type: () -> str
        return self.__client.account_id

    @property
    def auth_token(self):
        # type: () -> str
        return self.__client.auth_token

    @handle_request_exception
    def all(self):
        # type: () -> List[Comparison]
        return list(map(comparison_from_response, self.__client.get(self.__url)['results']))

    @handle_request_exception
    def get(self, identifier):
        # type: (str) -> Comparison
        identifier = validate_identifier(identifier)
        return comparison_from_response(self.__client.get(self.__url / identifier))

    @handle_request_exception
    def create(self, left, right, identifier=None, public=False, expires=None):
        """Creates a new comparison with the Draftable API.

        :param left: a string representing URL or file path, *or* a Side object that includes file type code and display name.
        :param right: as for "left".
        :param identifier: The identifier to use for this comparison, or None to generate a new identifier
        :param public: True if this comparison should be public, or False if not
        :param expires: None for never expires, or a datetime/timedelta object
        :return: the newly created comparison
        """
        # type: (Union[str, FileSide, URLSide], Union[str, FileSide, URLSide], Optional[str], bool, Optional[Union[datetime, timedelta]]) -> Comparison
        if identifier is not None:
            identifier = validate_identifier(identifier)
        if expires is not None:
            expires = validate_expires(expires)
        public = bool(public)

        data = {
            'identifier': identifier,
            'left': data_from_side('left', left),
            'right': data_from_side('right', right),
            'public': public,
            'expires': expires.isoformat() if expires is not None else None,
        }

        return comparison_from_response(self.__client.post(self.__url, data))

    @handle_request_exception
    def delete(self, identifier):
        # type: (str) -> None
        identifier = validate_identifier(identifier)
        self.__client.delete(self.__url / identifier)

    def public_viewer_url(self, identifier, wait=False):
        # type: (str, bool) -> str
        identifier = validate_identifier(identifier)
        return str(self.__url / 'viewer' / self.account_id / identifier + ('?wait' if wait else ''))

    def signed_viewer_url(self, identifier, valid_until=timedelta(minutes=30), wait=False):
        # type: (str, Union[datetime, timedelta], bool) -> str
        identifier = validate_identifier(identifier)
        valid_until_timestamp = aware_datetime_to_timestamp(validate_valid_until(valid_until))
        signature = signing.get_viewer_url_signature(self.account_id, self.auth_token, identifier, valid_until_timestamp)
        params = '?valid_until={valid_until}&signature={signature}{wait}'.format(
            valid_until=valid_until_timestamp,
            signature=signature,
            wait='&wait' if wait else '',
        )
        return str(self.__url / 'viewer' / self.account_id / identifier + params)



from __future__ import absolute_import

from datetime import datetime

from ...utilities.timestamp import parse_datetime

try:
    # noinspection PyUnresolvedReferences
    from typing import Optional
except ImportError:
    pass


class ComparisonSide(object):
    def __init__(self, file_type, source_url, display_name):
        # type: (str, Optional[str], Optional[str]) -> None
        self.__file_type = file_type
        self.__source_url = source_url
        self.__display_name = display_name

    @property
    def file_type(self):
        # type: () -> str
        return self.__file_type

    @property
    def source_url(self):
        # type: () -> Optional[str]
        return self.__source_url

    @property
    def display_name(self):
        # type: () -> Optional[str]
        return self.__display_name

    def __str__(self):
        # type: () -> str
        return '{{file_type: {}{}{}}}'.format(
            repr(self.file_type),
            ', source_url: {}'.format(repr(self.source_url)) if self.source_url else '',
            ', display_name: {}'.format(repr(self.display_name)) if self.display_name else '',
        )

    def __repr__(self):
        # type: () -> str
        return 'ComparisonSide(file_type={}, source_url={}, display_name={})'.format(
            repr(self.file_type),
            repr(self.source_url),
            repr(self.display_name),
        )


class Comparison(object):
    def __init__(self,
                 identifier,  # type: str
                 left,  # type: ComparisonSide
                 right,  # type: ComparisonSide
                 public,  # type: bool
                 creation_time,  # type: datetime
                 expiry_time,  # type: Optional[datetime]
                 ready,  # type: bool
                 ready_time,  # type: Optional[datetime]
                 failed,  # type: Optional[bool]
                 error_message  # type: Optional[str]
                 ):
        self.__identifier = identifier
        self.__left = left
        self.__right = right
        self.__creation_time = creation_time
        self.__public = public
        self.__expiry_time = expiry_time
        self.__ready = ready
        self.__ready_time = ready_time
        self.__failed = failed
        self.__error_message = error_message

    @property
    def identifier(self):
        # type: () -> str
        return self.__identifier

    @property
    def left(self):
        # type: () -> ComparisonSide
        return self.__left

    @property
    def right(self):
        # type: () -> ComparisonSide
        return self.__right

    @property
    def public(self):
        # type: () -> bool
        return self.__public

    @property
    def creation_time(self):
        # type: () -> datetime
        return self.__creation_time

    @property
    def expiry_time(self):
        # type: () -> Optional[datetime]
        return self.__expiry_time

    @property
    def ready(self):
        # type: () -> bool
        return self.__ready

    @property
    def ready_time(self):
        # type: () -> Optional[datetime]
        return self.__ready_time

    @property
    def failed(self):
        # type: () -> Optional[bool]
        return self.__failed

    @property
    def error_message(self):
        # type: () -> Optional[str]
        return self.__error_message

    def __str__(self):
        # type: () -> str
        return 'Comparison(identifier={}, left={}, right={}, public={}, creation_time={}, expiry_time={}, ready={}, ready_time={}, failed={}, error_message={})'.format(
            repr(self.identifier),
            str(self.left),
            str(self.right),
            str(self.public),
            self.creation_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
            self.expiry_time.strftime('%Y-%m-%dT%H:%M:%SZ') if self.expiry_time else None,
            str(self.ready),
            self.ready_time.strftime('%Y-%m-%dT%H:%M:%SZ') if self.ready_time else None,
            str(self.failed),
            '<{} chars>'.format(len(self.error_message)) if self.error_message is not None else None,
        )

    def __repr__(self):
        # type: () -> str
        return 'Comparison(identifier={}, left={}, right={}, public={}, creation_time={}, expiry_time={}, ready={}, ready_time={}, failed={}, error_message={})'.format(
            repr(self.identifier),
            repr(self.left),
            repr(self.right),
            repr(self.public),
            repr(self.creation_time),
            repr(self.expiry_time),
            repr(self.ready),
            repr(self.ready_time),
            repr(self.failed),
            repr(self.error_message),
        )


def _comparison_side_from_response(side_data):
    # type: (dict) -> ComparisonSide
    return ComparisonSide(file_type=str(side_data['file_type']), source_url=side_data.get('source_url'),
                          display_name=side_data.get('display_name'))


def comparison_from_response(data):
    # type: (dict) -> Comparison
    return Comparison(
        identifier=str(data['identifier']),
        left=_comparison_side_from_response(data['left']),
        right=_comparison_side_from_response(data['right']),
        public=data.get('public', False),
        creation_time=parse_datetime(data['creation_time']),
        expiry_time=parse_datetime(data['expiry_time']) if 'expiry_time' in data else None,
        ready=data.get('ready'),
        ready_time=parse_datetime(data['ready_time']) if 'ready_time' in data else None,
        failed=data.get('failed'),
        error_message=data.get('error_message'),
    )



from __future__ import absolute_import, division

try:
    # noinspection PyUnresolvedReferences
    from typing import Any, Union, Tuple, Optional
except ImportError:
    pass


class Url(object):
    """A pathlib-like object for building URLs.

    Join URLs using '/', and append query string using '+':

        >>> from draftable.endpoints.urls import Url
        >>> u = Url("http://example.com/foo")
        >>> u.full
        'http://example.com/foo'
        >>> u / "bar"
        Url('http://example.com/foo/bar')
        >>> u / "bar" / "baz.pdf"
        Url('http://example.com/foo/bar/baz.pdf')
        >>> u / "bar" / "baz.pdf" + "?t=1234"
        Url('http://example.com/foo/bar/baz.pdf?t=1234')
    """
    def __init__(self, base_url, *parts):
        # type: (Union[str, Url]) -> None
        self.__url = str(base_url) + self.to_append(*parts)  # the str(base_url) handles the case that `base_url` is already a Url

    def __str__(self):
        """Return this URL as a string."""
        return self.__url

    def __repr__(self):
        return "Url(%r)" % self.__url

    def __truediv__(self, key):
        """Append a single path item to this URL."""
        return Url(self.__url, key)

    def __add__(self, other):
        """Add directly to the URL. Suitable for appending query strings, for example."""
        return Url(self.__url + other)

    def __iadd__(self, other):
        self.__url += other
        return self

    def to_append(self, *parts):
        """Return the string of parts, joined by '/', if there are parts, or the empty string if no parts.

        The returned string is suitable for appending to a base URL.
        """
        return ('/' + '/'.join(parts)) if parts else ''

    @property
    def full(self):
        return self.__url
